import {
  Actions,
  StateContext,
  Store,
} from "@ngxs/store";
import {Observable} from "rxjs";
import {catchError} from "rxjs/internal/operators/catchError";
import {tap} from "rxjs/internal/operators/tap";
import {RegisterDefaultAction} from "../../decorators/store.decorator";
import {ApiResourceNameEnum} from "../../enums/partial/api-resource-name.enum";
import {urlSeparator} from "../../enums/partial/routes.enum";
import {ApiService} from "../../http/api.service";
import {BaseResourceType} from "../../models";
import {CollectionTyped} from "../../models/dto/collection-typed.model";
import {QueryParameters} from "../../models/query-parameters/query-parameters.model";
import {NotificationService} from "../../services";
import {StoreUtil} from "../../utils/stores/store.util";
import {
  BaseState,
  defaultBaseStateInitValue,
} from "../base/base.state";
import {AssociationNoSqlReadOnlyActionHelper} from "./association-no-sql-read-only-action.helper";
import {AssociationNoSqlReadOnlyNameSpace} from "./association-no-sql-read-only-namespace.model";
import {AssociationNoSqlReadOnlyOptions} from "./association-no-sql-read-only-options.model";
import {AssociationNoSqlReadOnlyStateModel} from "./association-no-sql-read-only-state.model";
import {AssociationNoSqlReadOnlyAction} from "./association-no-sql-read-only.action";

export const defaultAssociationNoSqlReadOnlyStateInitValue: () => AssociationNoSqlReadOnlyStateModel<BaseResourceType> = () =>
  ({
    ...defaultBaseStateInitValue(),
    total: 0,
    selected: undefined,
    current: undefined,
    queryParameters: new QueryParameters(),
  });

// @dynamic
export abstract class AssociationNoSqlReadOnlyState<TResource extends BaseResourceType> extends BaseState {
  protected readonly _resourceName: ApiResourceNameEnum;
  protected readonly _nameSpace: AssociationNoSqlReadOnlyNameSpace;

  protected constructor(protected apiService: ApiService,
                        protected store: Store,
                        protected notificationService: NotificationService,
                        protected actions$: Actions,
                        protected options: AssociationNoSqlReadOnlyOptions) {
    super(apiService, store, notificationService, actions$, options, AssociationNoSqlReadOnlyState);
    this._resourceName = this.options.resourceName;
  }

  private evaluateSubResourceUrl(parentId: string): string {
    return this._urlResource + urlSeparator + parentId + urlSeparator + this._resourceName;
  }

  // @InheritedSelector()
  // static current<TResource extends BaseResourceType>(state: AssociationNoSqlReadOnlyStateModel<BaseResourceType>): TResource {
  //   return state.current;
  // }
  //
  // @InheritedSelector()
  // static selected<TResource extends BaseResourceType>(state: AssociationNoSqlReadOnlyStateModel<BaseResourceType>): TResource[] {
  //   return state.selected;
  // }
  //
  // @InheritedSelector()
  // static total<T>(state: AssociationNoSqlReadOnlyStateModel<T>): number {
  //   return state.total;
  // }
  //
  // @InheritedSelector()
  // static queryParameters<T>(state: AssociationNoSqlReadOnlyStateModel<T>): QueryParameters {
  //   return state.queryParameters;
  // }

  @RegisterDefaultAction((associationNoSqlReadOnlyNameSpace: AssociationNoSqlReadOnlyNameSpace) => associationNoSqlReadOnlyNameSpace.GetAll)
  getAll<U>(ctx: StateContext<AssociationNoSqlReadOnlyStateModel<TResource>>, action: AssociationNoSqlReadOnlyAction.GetAll): Observable<CollectionTyped<U>> {
    let reset = {};
    if (!action.keepCurrentContext) {
      reset = {
        total: 0,
        selected: undefined,
      };
    }
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
      queryParameters: StoreUtil.getQueryParametersToApply(action.queryParameters, ctx),
      ...reset,
    });
    const url = this.evaluateSubResourceUrl(action.parentId);
    return this.apiService.get<U>(url, ctx.getState().queryParameters)
      .pipe(
        tap((collection: CollectionTyped<U>) => {
          ctx.dispatch(AssociationNoSqlReadOnlyActionHelper.getAllSuccess<U>(this._nameSpace, action, collection));
        }),
        catchError(error => {
          ctx.dispatch(AssociationNoSqlReadOnlyActionHelper.getAllFail(this._nameSpace, action));
          throw error;
        }),
      );
  }

  @RegisterDefaultAction((associationNoSqlReadOnlyNameSpace: AssociationNoSqlReadOnlyNameSpace) => associationNoSqlReadOnlyNameSpace.GetAllSuccess)
  getAllSuccess(ctx: StateContext<AssociationNoSqlReadOnlyStateModel<TResource>>, action: AssociationNoSqlReadOnlyAction.GetAllSuccess<TResource>): void {
    ctx.patchState({
      selected: action.list._data,
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((associationNoSqlReadOnlyNameSpace: AssociationNoSqlReadOnlyNameSpace) => associationNoSqlReadOnlyNameSpace.GetAllFail)
  getAllFail(ctx: StateContext<AssociationNoSqlReadOnlyStateModel<TResource>>, action: AssociationNoSqlReadOnlyAction.GetAllFail): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((associationNoSqlReadOnlyNameSpace: AssociationNoSqlReadOnlyNameSpace) => associationNoSqlReadOnlyNameSpace.GetById)
  getById<U>(ctx: StateContext<AssociationNoSqlReadOnlyStateModel<TResource>>, action: AssociationNoSqlReadOnlyAction.GetById): Observable<U> {
    let reset = {};
    if (!action.keepCurrentContext) {
      reset = {
        current: undefined,
      };
    }
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
      ...reset,
    });
    const url = this.evaluateSubResourceUrl(action.parentId);
    return this.apiService.getById<U>(url, action.resId)
      .pipe(
        tap((model: U) => {
          ctx.dispatch(AssociationNoSqlReadOnlyActionHelper.getByIdSuccess<U>(this._nameSpace, action, model));
        }),
        catchError(error => {
          ctx.dispatch(AssociationNoSqlReadOnlyActionHelper.getByIdFail(this._nameSpace, action));
          throw error;
        }),
      );
  }

  @RegisterDefaultAction((associationNoSqlReadOnlyNameSpace: AssociationNoSqlReadOnlyNameSpace) => associationNoSqlReadOnlyNameSpace.GetByIdSuccess)
  getByIdSuccess(ctx: StateContext<AssociationNoSqlReadOnlyStateModel<TResource>>, action: AssociationNoSqlReadOnlyAction.GetByIdSuccess<TResource>): void {
    ctx.patchState({
      current: action.model,
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((associationNoSqlReadOnlyNameSpace: AssociationNoSqlReadOnlyNameSpace) => associationNoSqlReadOnlyNameSpace.GetByIdFail)
  getByIdFail(ctx: StateContext<AssociationNoSqlReadOnlyStateModel<TResource>>, action: AssociationNoSqlReadOnlyAction.GetByIdFail): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }
}
