import {
  Actions,
  StateContext,
  Store,
} from "@ngxs/store";
import {Observable} from "rxjs";
import {catchError} from "rxjs/internal/operators/catchError";
import {tap} from "rxjs/internal/operators/tap";
import {RegisterDefaultAction} from "../../decorators/store.decorator";
import {ApiService} from "../../http/api.service";
import {BaseResourceType} from "../../models";
import {CollectionTyped} from "../../models/dto/collection-typed.model";
import {QueryParameters} from "../../models/query-parameters/query-parameters.model";
import {NotificationService} from "../../services";
import {StoreUtil} from "../../utils/stores/store.util";
import {
  BaseState,
  defaultBaseStateInitValue,
} from "../base/base.state";
import {ResourceNoSqlReadOnlyActionHelper} from "./resource-no-sql-read-only-action.helper";
import {ResourceNoSqlReadOnlyNameSpace} from "./resource-no-sql-read-only-namespace.model";
import {ResourceNoSqlReadOnlyOptions} from "./resource-no-sql-read-only-options.model";
import {ResourceNoSqlReadOnlyStateModel} from "./resource-no-sql-read-only-state.model";
import {ResourceNoSqlReadOnlyAction} from "./resource-no-sql-read-only.action";

export const defaultResourceNoSqlReadOnlyStateInitValue: () => ResourceNoSqlReadOnlyStateModel<BaseResourceType> = () =>
  ({
    ...defaultBaseStateInitValue(),
    total: 0,
    list: undefined,
    current: undefined,
    queryParameters: new QueryParameters(),
  });

// @dynamic
export abstract class ResourceNoSqlReadOnlyState<TResource extends BaseResourceType> extends BaseState {
  protected readonly _nameSpace: ResourceNoSqlReadOnlyNameSpace;

  protected constructor(protected apiService: ApiService,
                        protected store: Store,
                        protected notificationService: NotificationService,
                        protected actions$: Actions,
                        protected options: ResourceNoSqlReadOnlyOptions) {
    super(apiService, store, notificationService, actions$, options, ResourceNoSqlReadOnlyState);
  }

  // @InheritedSelector()
  // static current<TResource extends BaseResourceType>(state: ResourceNoSqlReadOnlyStateModel<BaseResourceType>): TResource {
  //   return state.current;
  // }
  //
  // @InheritedSelector()
  // static list<T>(state: ResourceNoSqlReadOnlyStateModel<T>): T[] {
  //   return state.list;
  // }
  //
  // @InheritedSelector()
  // static total<T>(state: ResourceNoSqlReadOnlyStateModel<T>): number {
  //   return state.total;
  // }
  //
  // @InheritedSelector()
  // static queryParameters<T>(state: ResourceNoSqlReadOnlyStateModel<T>): QueryParameters {
  //   return state.queryParameters;
  // }

  @RegisterDefaultAction((resourceNoSqlReadOnlyNameSpace: ResourceNoSqlReadOnlyNameSpace) => resourceNoSqlReadOnlyNameSpace.GetAll)
  getAll(ctx: StateContext<ResourceNoSqlReadOnlyStateModel<TResource>>, action: ResourceNoSqlReadOnlyAction.GetAll): Observable<CollectionTyped<TResource>> {
    let reset = {};
    if (!action.keepCurrentContext) {
      reset = {
        total: 0,
        list: undefined,
      };
    }
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
      queryParameters: StoreUtil.getQueryParametersToApply(action.queryParameters, ctx),
      ...reset,
    });
    return this.apiService.get<TResource>(this._urlResource, ctx.getState().queryParameters)
      .pipe(
        tap((collection: CollectionTyped<TResource>) => {
          ctx.dispatch(ResourceNoSqlReadOnlyActionHelper.getAllSuccess<TResource>(this._nameSpace, action, collection));
        }),
        catchError(error => {
          ctx.dispatch(ResourceNoSqlReadOnlyActionHelper.getAllFail(this._nameSpace, action));
          throw error;
        }),
      );
  }

  @RegisterDefaultAction((resourceNoSqlReadOnlyNameSpace: ResourceNoSqlReadOnlyNameSpace) => resourceNoSqlReadOnlyNameSpace.GetAllSuccess)
  getAllSuccess(ctx: StateContext<ResourceNoSqlReadOnlyStateModel<TResource>>, action: ResourceNoSqlReadOnlyAction.GetAllSuccess<TResource>): void {
    ctx.patchState({
      list: action.list._data,
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((resourceNoSqlReadOnlyNameSpace: ResourceNoSqlReadOnlyNameSpace) => resourceNoSqlReadOnlyNameSpace.GetAllFail)
  getAllFail(ctx: StateContext<ResourceNoSqlReadOnlyStateModel<TResource>>, action: ResourceNoSqlReadOnlyAction.GetAllFail): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((resourceNoSqlReadOnlyNameSpace: ResourceNoSqlReadOnlyNameSpace) => resourceNoSqlReadOnlyNameSpace.GetById)
  getById(ctx: StateContext<ResourceNoSqlReadOnlyStateModel<TResource>>, action: ResourceNoSqlReadOnlyAction.GetById): Observable<TResource> {
    let reset = {};
    if (!action.keepCurrentContext) {
      reset = {
        current: undefined,
      };
    }
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
      ...reset,
    });

    return this.apiService.getById<TResource>(this._urlResource, action.id)
      .pipe(
        tap((model: TResource) => {
          ctx.dispatch(ResourceNoSqlReadOnlyActionHelper.getByIdSuccess(this._nameSpace, action, model));
        }),
        catchError(error => {
          ctx.dispatch(ResourceNoSqlReadOnlyActionHelper.getByIdFail(this._nameSpace, action));
          throw error;
        }),
      );
  }

  @RegisterDefaultAction((resourceNoSqlReadOnlyNameSpace: ResourceNoSqlReadOnlyNameSpace) => resourceNoSqlReadOnlyNameSpace.GetByIdSuccess)
  getByIdSuccess(ctx: StateContext<ResourceNoSqlReadOnlyStateModel<TResource>>, action: ResourceNoSqlReadOnlyAction.GetByIdSuccess<TResource>): void {
    ctx.patchState({
      current: action.model,
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((resourceNoSqlReadOnlyNameSpace: ResourceNoSqlReadOnlyNameSpace) => resourceNoSqlReadOnlyNameSpace.GetByIdFail)
  getByIdFail(ctx: StateContext<ResourceNoSqlReadOnlyStateModel<TResource>>, action: ResourceNoSqlReadOnlyAction.GetByIdFail): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }
}
