import {Navigate} from "@ngxs/router-plugin";
import {
  Actions,
  StateContext,
  Store,
} from "@ngxs/store";
import {Observable} from "rxjs";
import {catchError} from "rxjs/internal/operators/catchError";
import {tap} from "rxjs/internal/operators/tap";
import {RegisterDefaultAction} from "../../decorators";
import {ModelAttributeEnum} from "../../enums";
import {ApiService} from "../../http/api.service";
import {
  BaseResourceType,
  CollectionTyped,
  QueryParameters,
} from "../../models";
import {NotificationService} from "../../services";
import {
  isFalse,
  isNullOrUndefined,
  isTrue,
} from "../../tools";
import {StoreUtil} from "../../utils/stores/store.util";
import {BaseOptions} from "../base/base-options.model";
import {
  BaseState,
  defaultBaseStateInitValue,
} from "../base/base.state";
import {ResourceActionHelper} from "./resource-action.helper";
import {ResourceNameSpace} from "./resource-namespace.model";
import {ResourceOptions} from "./resource-options.model";
import {ResourceStateModel} from "./resource-state.model";
import {ResourceAction} from "./resource.action";

export const defaultResourceStateInitValue: () => ResourceStateModel<BaseResourceType> = () =>
  ({
    ...defaultBaseStateInitValue(),
    total: 0,
    list: undefined,
    current: undefined,
    queryParameters: new QueryParameters(),
  });

// @dynamic
export abstract class ResourceState<TResource extends BaseResourceType> extends BaseState {
  protected readonly _nameSpace: ResourceNameSpace;

  protected constructor(protected apiService: ApiService,
                        protected store: Store,
                        protected notificationService: NotificationService,
                        protected actions$: Actions,
                        protected options: ResourceOptions) {
    super(apiService, store, notificationService, actions$, options, ResourceState);
  }

  protected static getDefaultOptions(): ResourceOptions | any {
    // TODO To fix, seems to not work
    let defaultOptions: ResourceOptions | any = {
      keepCurrentStateAfterCreate: false,
      keepCurrentStateAfterUpdate: false,
      keepCurrentStateAfterDelete: false,
    };
    defaultOptions = Object.assign(BaseState.getDefaultOptions(), defaultOptions);
    return defaultOptions;
  }

  protected static checkOptions(stateName: string, options: BaseOptions): void {
    BaseState.checkOptions(stateName, options);
  }

  // @InheritedSelector()
  // static current<TResource extends BaseResourceType>(state: ResourceStateModel<BaseResourceType>): TResource {
  //   return state.current;
  // }
  //
  // @InheritedSelector()
  // static list<T>(state: ResourceStateModel<T>): T[] {
  //   return state.list;
  // }
  //
  // @InheritedSelector()
  // static total<T>(state: ResourceStateModel<T>): number {
  //   return state.total;
  // }
  //
  // @InheritedSelector()
  // static queryParameters<T>(state: ResourceStateModel<T>): QueryParameters {
  //   return state.queryParameters;
  // }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.LoadResourceSuccess)
  loadResourceSuccess(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.LoadResourceSuccess): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.LoadResourceFail)
  loadResourceFail(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.LoadResourceFail): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.ChangeQueryParameters)
  changeQueryParameters(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.ChangeQueryParameters): void {
    ctx.patchState({
      queryParameters: action.queryParameters,
    });
    ctx.dispatch(ResourceActionHelper.getAll(this._nameSpace, undefined, action.keepCurrentContext));
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.GetAll)
  getAll(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.GetAll): Observable<CollectionTyped<TResource>> {
    let reset = {};
    if (!action.keepCurrentContext) {
      reset = {
        list: undefined,
        total: 0,
      };
    }
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
      queryParameters: StoreUtil.getQueryParametersToApply(action.queryParameters, ctx),
      ...reset,
    });
    return this.apiService.get<TResource>(this._urlResource, ctx.getState().queryParameters)
      .pipe(
        StoreUtil.cancelUncompleted(ctx, this.actions$, [this._nameSpace.GetAll, Navigate]),
        tap((collection: CollectionTyped<TResource>) => {
          ctx.dispatch(ResourceActionHelper.getAllSuccess<TResource>(this._nameSpace, action, collection));
        }),
        catchError(error => {
          ctx.dispatch(ResourceActionHelper.getAllFail(this._nameSpace, action));
          throw error;
        }),
      );
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.GetAllSuccess)
  getAllSuccess(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.GetAllSuccess<TResource>): void {
    const queryParameters = StoreUtil.updateQueryParameters(ctx, action.list);

    ctx.patchState({
      list: action.list._data,
      total: action.list._page.totalItems,
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
      queryParameters,
    });
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.GetAllFail)
  getAllFail(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.GetAllFail<TResource>): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.GetById)
  getById(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.GetById): Observable<TResource> {
    let reset = {};
    if (!action.keepCurrentContext) {
      reset = {
        current: undefined,
      };
    }
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
      ...reset,
    });

    return this.apiService.getById<TResource>(this._urlResource, action.id)
      .pipe(
        tap((model: TResource) => {
          ctx.dispatch(ResourceActionHelper.getByIdSuccess(this._nameSpace, action, model));
        }),
        catchError(error => {
          ctx.dispatch(ResourceActionHelper.getByIdFail(this._nameSpace, action));
          throw error;
        }),
      );
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.GetByIdSuccess)
  getByIdSuccess(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.GetByIdSuccess<TResource>): void {
    ctx.patchState({
      current: action.model,
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.GetByIdFail)
  getByIdFail(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.GetByIdFail<TResource>): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.Create)
  create(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.Create<TResource>): Observable<TResource> {
    return this.internalCreate(ctx, action)
      .pipe(
        tap((model: TResource) => {
          ctx.dispatch(ResourceActionHelper.createSuccess(this._nameSpace, action, model));
        }),
      );
  }

  protected internalCreate(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.Create<TResource>): Observable<TResource> {
    this.cleanCurrentStateIfDefined(ctx, this.options.keepCurrentStateBeforeCreate);
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
    });

    return this.apiService.post<TResource>(this._urlResource, action.modelFormControlEvent.model)
      .pipe(
        catchError(error => {
          ctx.dispatch(ResourceActionHelper.createFail(this._nameSpace, action));
          throw error;
        }),
        StoreUtil.catchValidationErrors(ctx, action.modelFormControlEvent),
      );
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.CreateSuccess)
  createSuccess(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.CreateSuccess<TResource>): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
      current: action.model,
    });

    StoreUtil.navigateIfDefined(ctx, this.options.routeRedirectUrlAfterSuccessCreateAction, action.model.resId);
    StoreUtil.notifySuccessIfDefined(this.notificationService, this.options.notificationResourceCreateTextToTranslate);
    this.cleanCurrentStateIfDefined(ctx, this.options.keepCurrentStateAfterCreate);
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.CreateFail)
  createFail(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.CreateFail<TResource>): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.Update)
  update(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.Update<TResource>): Observable<TResource> {
    return this.internalUpdate(ctx, action)
      .pipe(
        tap((result) => {
          ctx.dispatch(ResourceActionHelper.updateSuccess(this._nameSpace, action, result));
        }),
      );
  }

  protected internalUpdate(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.Update<TResource>): Observable<TResource> {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
    });

    return this.apiService.patchById<TResource>(this._urlResource, action.modelFormControlEvent.model[ModelAttributeEnum.resId], action.modelFormControlEvent.model)
      .pipe(
        catchError(error => {
          ctx.dispatch(ResourceActionHelper.updateFail(this._nameSpace, action));
          throw error;
        }),
        StoreUtil.catchValidationErrors(ctx, action.modelFormControlEvent),
      );
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.UpdateSuccess)
  updateSuccess(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.UpdateSuccess<TResource>): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
      current: action.model,
    });

    StoreUtil.navigateIfDefined(ctx, this.options.routeRedirectUrlAfterSuccessUpdateAction, action.model.resId);
    StoreUtil.notifySuccessIfDefined(this.notificationService, this.options.notificationResourceUpdateTextToTranslate);
    this.cleanCurrentStateIfDefined(ctx, this.options.keepCurrentStateAfterUpdate);
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.UpdateFail)
  updateFail(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.UpdateFail<TResource>): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.Delete)
  delete(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.Delete): Observable<TResource> {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
    });

    return this.apiService.deleteById<TResource>(this._urlResource, action.resId)
      .pipe(
        tap(() => {
          ctx.dispatch(ResourceActionHelper.deleteSuccess(this._nameSpace, action));
        }),
        catchError(error => {
          ctx.dispatch(ResourceActionHelper.deleteFail(this._nameSpace, action));
          throw error;
        }),
      );
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.DeleteSuccess)
  deleteSuccess(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.DeleteSuccess): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });

    StoreUtil.navigateIfDefined(ctx, this.options.routeRedirectUrlAfterSuccessDeleteAction, undefined);
    StoreUtil.notifySuccessIfDefined(this.notificationService, this.options.notificationResourceDeleteTextToTranslate);
    this.cleanCurrentStateIfDefined(ctx, this.options.keepCurrentStateAfterDelete);
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.DeleteFail)
  deleteFail(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.DeleteFail): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((resourceNameSpace: ResourceNameSpace) => resourceNameSpace.Clean)
  clean(ctx: StateContext<ResourceStateModel<TResource>>, action: ResourceAction.Clean): void {
    this.resetStateToDefault(ctx, true);
  }

  private cleanCurrentStateIfDefined(ctx: StateContext<ResourceStateModel<TResource>>, keepCurrentState: boolean): void {
    if (isNullOrUndefined(keepCurrentState) || isFalse(keepCurrentState)) {
      this.resetStateToDefault(ctx, true);
    }
  }

  private resetStateToDefault(ctx: StateContext<ResourceStateModel<TResource>>, preserveListData: boolean): ResourceStateModel<BaseResourceType> {
    const list = ctx.getState().list;
    const total = ctx.getState().total;
    const queryParameters = ctx.getState().queryParameters;
    let result = ctx.setState(this.getDefaultData());
    if (isTrue(preserveListData)) {
      result = ctx.patchState({
        list,
        total,
        queryParameters,
      });
    }
    return result;
  }

  private getDefaultData(): ResourceStateModel<BaseResourceType> {
    return StoreUtil.getDefaultData(this.constructor);
  }
}
